@@ -4,11 +4,13 @@ import os.path
 import posixpath
 import shutil
 from dataclasses import dataclass
+from pathlib import Path
 from typing import Any, Dict, Iterable, Iterator, List, Optional, Tuple
 
 import yaml
 from dvc_data.hashfile.db.local import LocalHashFileDB
-from dvc_objects.fs.local import LocalFileSystem
+from dvc_objects.fs.local import LocalFileSystem as _LocalFileSystem
+from fsspec.implementations.local import LocalFileSystem
 from tqdm import tqdm
 
 from dql.client import Client
@@ -37,10 +39,10 @@ class DataSource:
         self.node = node
 
     def get_full_path(self):
-        return posixpath.join(self.listing.storage.uri, self.node.full_path)
+        return self.get_node_full_path(self.node)
 
     def get_node_full_path(self, node):
-        return posixpath.join(self.listing.storage.uri, node.full_path)
+        return self.listing.client.get_full_path(node.full_path)
 
 
 @dataclass
@@ -311,7 +313,7 @@ def find_column_to_str(node: NodeWithPath, src: DataSource, column: str) -> str:
     if column == "owner":
         return node.owner_name or ""
     if column == "path":
-        return posixpath.join(src.get_full_path(), node.full_path)
+        return src.get_node_full_path(node)
     if column == "size":
         return str(node.size)
     if column == "type":
@@ -329,11 +331,23 @@ class Catalog:
         dql_dir = DQLDir(cache=cache_dir, tmp=tmp_dir)
         self.data_storage = data_storage
         self.cache = LocalHashFileDB(
-            LocalFileSystem(),
+            _LocalFileSystem(),
             dql_dir.cache,
             tmp_dir=dql_dir.tmp,
         )
 
+    def add_storage(self, source: str) -> None:
+        # pylint:disable-next=protected-access
+        path = LocalFileSystem._strip_protocol(source)
+        if not os.path.isdir(path):
+            raise RuntimeError(f"{path} is not a directory")
+        uri = Path(path).as_uri()
+        print(f"Registering storage {uri}")
+        self.data_storage.create_storage_if_not_registered(uri)
+
+    def parse_url(self, uri: str, **config: Any) -> Tuple[Client, str]:
+        return Client.parse_url(uri, self.data_storage, **config)
+
     def enlist_source(
         self,
         source: str,
@@ -350,7 +364,7 @@ class Catalog:
             )
 
         client_config = client_config or {}
-        client, path = Client.parse_url(source, **client_config)
+        client, path = self.parse_url(source, **client_config)
         prefix = posixpath.dirname(path)
         source_data_storage = self.data_storage.clone(uri=client.uri)
 
@@ -502,7 +516,7 @@ class Catalog:
         client_config = client_config or {}
         self.data_storage.create_shadow_dataset(name)
         for source in sources:
-            client, path = Client.parse_url(source, **client_config)
+            client, path = self.parse_url(source, **client_config)
             self.data_storage.insert_into_shadow_dataset(
                 name, client.uri, path, recursive=recursive
             )
@@ -515,15 +529,16 @@ class Catalog:
         description: Optional[str] = None,
         labels: Optional[List[str]] = None,
     ):
+        version = version or 1
+
         dataset = self.data_storage.get_dataset(dataset_name)
         if not dataset:
-            raise RuntimeError(f"Dataset with name: {dataset_name} doesn't exist")
+            raise ValueError(f"Dataset {dataset_name} doesn't exist")
 
         if not dataset.shadow:
-            raise RuntimeError(f"Dataset is already registered: {dataset_name}")
+            raise ValueError(f"Dataset {dataset_name} is already registered")
 
         update_data = {
-            "version": version or 1,
             "shadow": False,
             "description": description,
             "labels": labels,
@@ -533,15 +548,52 @@ class Catalog:
             update_data["name"] = new_dataset_name
 
         self.data_storage.update_dataset(dataset_name, **update_data)
+        self.data_storage.create_dataset_version(
+            new_dataset_name or dataset_name, version, create_rows_table=False
+        )
+        self.data_storage.rename_dataset_rows(new_dataset_name or dataset_name, version)
 
     def ls_datasets(self, shadow_only=None) -> Iterable[DatasetRecord]:
         yield from self.data_storage.get_datasets(shadow_only=shadow_only)
 
-    def ls_dataset_rows(self, name: str, limit=None) -> Iterable[DatasetRow]:
-        yield from self.data_storage.get_dataset_rows(name, limit=limit)
+    def ls_dataset_rows(
+        self, name: str, limit=None, version=None
+    ) -> Iterable[DatasetRow]:
+        dataset = self.data_storage.get_dataset(name)
+        assert dataset
+        if not dataset.shadow and not version:
+            raise ValueError(
+                f"Missing dataset version from input for registered dataset {name}"
+            )
 
-    def remove_dataset(self, name: str):
-        self.data_storage.remove_dataset(name)
+        yield from self.data_storage.get_dataset_rows(
+            name, limit=limit, version=version
+        )
+
+    def remove_dataset(
+        self,
+        name: str,
+        version: Optional[int] = None,
+        force: Optional[bool] = False,
+    ):
+        dataset = self.data_storage.get_dataset(name)
+        assert dataset
+        if dataset.registered and not version and not force:
+            raise ValueError(
+                f"Missing dataset version from input for registered dataset {name}"
+            )
+        if version and not dataset.has_version(version):
+            raise RuntimeError(f"Dataset {name} doesn't have version {version}")
+
+        if dataset.registered and version:
+            # removing one version of registered dataset
+            self.data_storage.remove_dataset_version(dataset, version)
+
+        elif dataset.registered and force:
+            for version in dataset.versions.copy():  # type: ignore [union-attr]
+                self.data_storage.remove_dataset_version(dataset, version)
+        else:
+            self.data_storage.remove_shadow_dataset(dataset)
 
     def edit_dataset(
         self,
@@ -560,6 +612,64 @@ class Catalog:
 
         self.data_storage.update_dataset(name, **update_data)
 
+    def merge_datasets(
+        self,
+        src_name: str,
+        dst_name: str,
+        src_version: Optional[int] = None,
+        dst_version: Optional[int] = None,
+    ) -> DatasetRecord:
+        """
+        Merges records from source to destination dataset.
+        If destination dataset is shadow, it will copy all the records from source
+        dataset to shadow one
+        If destination dataset is registered, it will create a new version
+        of dataset with records merged from old version and the source
+        """
+
+        src = self.data_storage.get_dataset(src_name)
+        dst = self.data_storage.get_dataset(dst_name)
+
+        # validation
+        if not src:
+            raise ValueError(f"Source dataset {src_name} doesn't exist")
+        if src.shadow and src_version:
+            raise ValueError(
+                f"Source dataset {src_name} is shadow, cannot use it with versions"
+            )
+        if not src.shadow and not src_version:
+            raise ValueError(f"Source dataset {src_name} is registered, need a version")
+        if not dst:
+            raise ValueError(f"Dataset {dst_name} doesn't exist")
+        if dst.shadow and dst_version:
+            raise ValueError(
+                f"Dataset {dst_name} is shadow, cannot use it with versions"
+            )
+
+        if dst_version and not dst.is_valid_next_version(dst_version):
+            raise ValueError(
+                f"Version {dst_version} must be higher than the current latest one"
+            )
+
+        if dst.shadow:
+            self.data_storage.merge_dataset_rows(
+                src,
+                dst,
+                src_version,
+                dst_version=None,
+            )
+        else:
+            dst_version = dst_version or dst.next_version
+            dst = self.data_storage.create_dataset_version(dst_name, dst_version)
+            self.data_storage.merge_dataset_rows(
+                src,
+                dst,
+                src_version,
+                dst_version,
+            )
+
+        return dst
+
     def ls(
         self,
         sources: List[str],
@@ -769,7 +879,7 @@ class Catalog:
         # saving them as storages, without further indexing in each bucket
         for source in root_sources:
             for bucket in Client.get_implementation(source).ls_buckets(**client_config):
-                client, _ = Client.parse_url(bucket.uri)
+                client, _ = self.parse_url(bucket.uri)
                 print(f"Registering storage {client.uri}")
                 self.data_storage.create_storage_if_not_registered(client.uri)
 
@@ -779,3 +889,6 @@ class Catalog:
             update,
             client_config=client_config,
         )
+
+    def find_stale_storages(self) -> None:
+        self.data_storage.find_stale_storages()
