@@ -239,11 +239,61 @@ def get_parser() -> ArgumentParser:
         "ls-dataset-rows", parents=[parent_parser], help="List dataset rows"
     )
     ls_dataset_rows_parser.add_argument("name", type=str, help="Dataset name")
+    ls_dataset_rows_parser.add_argument(
+        "--version",
+        action="store",
+        default=None,
+        type=int,
+        help="Dataset version",
+    )
 
     rm_dataset_parser = subp.add_parser(
         "rm-dataset", parents=[parent_parser], help="Removes dataset"
     )
     rm_dataset_parser.add_argument("name", type=str, help="Dataset name")
+    rm_dataset_parser.add_argument(
+        "--version",
+        action="store",
+        default=None,
+        type=int,
+        help="Dataset version",
+    )
+    rm_dataset_parser.add_argument(
+        "--force",
+        default=False,
+        action=BooleanOptionalAction,
+        help="Force delete registered dataset with all of it's versions",
+    )
+
+    parse_merge_datasets = subp.add_parser(
+        "merge-datasets", parents=[parent_parser], help="Merges datasets"
+    )
+    parse_merge_datasets.add_argument(
+        "--src",
+        action="store",
+        default=None,
+        help="Source dataset name",
+    )
+    parse_merge_datasets.add_argument(
+        "--dst",
+        action="store",
+        default=None,
+        help="Destination dataset name",
+    )
+    parse_merge_datasets.add_argument(
+        "--src-version",
+        action="store",
+        default=None,
+        type=int,
+        help="Source dataset version",
+    )
+    parse_merge_datasets.add_argument(
+        "--dst-version",
+        action="store",
+        default=None,
+        type=int,
+        help="Destination dataset version",
+    )
 
     parse_ls = subp.add_parser(
         "ls", parents=[parent_parser], help="List storage contents"
@@ -363,6 +413,17 @@ def get_parser() -> ArgumentParser:
         "index", parents=[parent_parser], help="Index storage location"
     )
     add_sources_arg(parse_index)
+    add_storage_parser = subp.add_parser(
+        "add-storage", parents=[parent_parser], help="Register local path as storage"
+    )
+    add_storage_parser.add_argument("uri", type=str, help="Source URI")
+
+    subp.add_parser(
+        "find-stale-storages",
+        parents=[parent_parser],
+        help="Finds and marks stale storages",
+    )
+
     add_completion_parser(subp, [parent_parser])
     return parser
 
@@ -415,6 +476,12 @@ def edit_dataset(name, **kwargs):
     catalog.edit_dataset(name, **kwargs)
 
 
+def merge_datasets(src, dst, **kwargs):
+    data_storage = SQLiteDataStorage()
+    catalog = Catalog(data_storage)
+    catalog.merge_datasets(src, dst, **kwargs)
+
+
 def ls_urls(
     sources,
     long: bool = False,
@@ -582,17 +649,18 @@ def ls_datasets(shadow_only=None):
         if d.shadow:
             print(f"(tmp) {d.name}")
         else:
-            print(d.name)
+            for v in d.versions or []:
+                print(f"{d.name} (v{v})")
 
 
-def ls_dataset_rows(name: str):
-    for row in Catalog(SQLiteDataStorage()).ls_dataset_rows(name):
+def ls_dataset_rows(name: str, version: Optional[int] = None):
+    for row in Catalog(SQLiteDataStorage()).ls_dataset_rows(name, version=version):
         entry = row.name + ("/" if row.dir_type else "")  # type: ignore[attr-defined]
         print(format_ls_entry(entry))
 
 
-def rm_dataset(name: str):
-    Catalog(SQLiteDataStorage()).remove_dataset(name)
+def rm_dataset(name: str, version: Optional[int] = None, force: Optional[bool] = False):
+    Catalog(SQLiteDataStorage()).remove_dataset(name, version=version, force=force)
 
 
 def du(sources, show_bytes=False, si=False, **kwargs):
@@ -617,6 +685,14 @@ def index(sources, catalog: Optional[Catalog] = None, **kwargs):
     catalog.index(sources, **kwargs)
 
 
+def find_stale_storages():
+    Catalog(SQLiteDataStorage()).find_stale_storages()
+
+
+def add_storage(uri: str):
+    Catalog(SQLiteDataStorage()).add_storage(uri)
+
+
 def completion(shell: str) -> str:
     return shtab.complete(
         get_parser(),
@@ -694,6 +770,13 @@ def main(argv: Optional[List[str]] = None) -> int:  # noqa: C901
                 new_name=args.new_name,
                 labels=args.labels,
             )
+        elif args.command == "merge-datasets":
+            merge_datasets(
+                args.src,
+                args.dst,
+                src_version=args.src_version,
+                dst_version=args.dst_version,
+            )
         elif args.command == "ls":
             ls(
                 args.sources,
@@ -706,9 +789,9 @@ def main(argv: Optional[List[str]] = None) -> int:  # noqa: C901
         elif args.command == "ls-datasets":
             ls_datasets(args.shadow)
         elif args.command == "ls-dataset-rows":
-            ls_dataset_rows(args.name)
+            ls_dataset_rows(args.name, args.version)
         elif args.command == "rm-dataset":
-            rm_dataset(args.name)
+            rm_dataset(args.name, version=args.version, force=args.force)
         elif args.command == "du":
             du(
                 args.sources,
@@ -746,8 +829,12 @@ def main(argv: Optional[List[str]] = None) -> int:  # noqa: C901
                 update=bool(args.update),
                 client_config=client_config,
             )
+        elif args.command == "add-storage":
+            add_storage(args.uri)
         elif args.command == "completion":
             print(completion(args.shell))
+        elif args.command == "find-stale-storages":
+            find_stale_storages()
         else:
             print(f"invalid command: {args.command}", file=sys.stderr)
             return 1
