@@ -8,11 +8,7 @@ class AzureClient(FSSpecClient):
     PREFIX = "az://"
     protocol = "az"
 
-    def __init__(self, name, **kwargs):
-        super().__init__(name, self.create_fs(**kwargs))
-
-    @classmethod
-    def _dict_from_info(cls, v, parent_id, delimiter, path):
+    def _dict_from_info(self, v, parent_id, delimiter, path):
         version_id = v.get("version_id")
         checksum = v.get("content_settings", {}).get("content_md5", "")
         name = v.get("name", "").split(delimiter)[-1]
