@@ -1,6 +1,9 @@
 from abc import ABC, abstractmethod
 from datetime import datetime
-from typing import ClassVar, Iterator, NamedTuple, Optional, Tuple, Type
+from typing import TYPE_CHECKING, ClassVar, Iterator, NamedTuple, Optional, Tuple, Type
+
+if TYPE_CHECKING:
+    from dql.data_storage import AbstractDataStorage
 
 
 class Bucket(NamedTuple):
@@ -17,6 +20,7 @@ class Client(ABC):
     def get_implementation(url: str) -> Type["Client"]:
         from .azure import AzureClient
         from .gcs import GCSClient
+        from .local import FileClient
         from .s3 import ClientS3
 
         if url.startswith(ClientS3.PREFIX):
@@ -25,12 +29,23 @@ class Client(ABC):
             return GCSClient
         elif url.startswith(AzureClient.PREFIX):
             return AzureClient
+        elif url.startswith(FileClient.PREFIX):
+            return FileClient
         raise RuntimeError(f"Unsupported data source format '{url}'")
 
     @staticmethod
-    def parse_url(source: str, **kwargs) -> Tuple["Client", str]:
+    def parse_url(
+        source: str, data_storage: "AbstractDataStorage", **kwargs
+    ) -> Tuple["Client", str]:
         cls = Client.get_implementation(source)
-        return cls._parse_url(source, **kwargs)  # pylint:disable=protected-access
+        storage, path = cls.split_url(source, data_storage)
+        client = cls.from_url(storage, **kwargs)
+        return client, path
+
+    @classmethod
+    @abstractmethod
+    def from_url(cls, url: str, **kwargs) -> "Client":
+        ...
 
     @classmethod
     @abstractmethod
@@ -39,12 +54,9 @@ class Client(ABC):
 
     @classmethod
     @abstractmethod
-    def split_url(cls, url: str) -> Tuple[str, str]:
-        ...
-
-    @classmethod
-    @abstractmethod
-    def _parse_url(cls, source: str, **kwargs) -> Tuple["Client", str]:
+    def split_url(
+        cls, url: str, data_storage: "AbstractDataStorage"
+    ) -> Tuple[str, str]:
         ...
 
     @classmethod
