@@ -25,7 +25,7 @@ class FSSpecClient(Client):
     FS_CLASS: ClassVar[Type["AbstractFileSystem"]]
     PREFIX: ClassVar[str]
 
-    def __init__(self, name, fs):
+    def __init__(self, name: str, fs: "AbstractFileSystem") -> None:
         self.name = name
         self.fs = fs
 
@@ -36,6 +36,10 @@ class FSSpecClient(Client):
         fs.invalidate_cache()
         return fs
 
+    @classmethod
+    def from_url(cls, url: str, **kwargs):
+        return cls(url, cls.create_fs(**kwargs))
+
     @classmethod
     def ls_buckets(cls, **kwargs) -> Iterator[Bucket]:
         for entry in cls.create_fs(**kwargs).ls(cls.PREFIX, detail=True):
@@ -55,29 +59,13 @@ class FSSpecClient(Client):
         return f"{self.PREFIX}{self.name}"
 
     @classmethod
-    def split_url(cls, url: str) -> Tuple[str, str]:
+    def split_url(cls, url: str, data_storage) -> Tuple[str, str]:
         fill_path = url[len(cls.PREFIX) :]
         path_split = fill_path.split("/", 1)
         bucket = path_split[0]
         path = path_split[1] if len(path_split) > 1 else ""
         return bucket, path
 
-    @classmethod
-    def _parse_url(
-        cls,
-        source: str,
-        **kwargs,
-    ) -> Tuple["FSSpecClient", str]:
-        """
-        Returns storage representation of bucket and the rest of the path
-        in source.
-        E.g if the source is s3://bucket_name/dir1/dir2/dir3 this method would
-        return Storage object of bucket_name and a path which is dir1/dir2/dir3
-        """
-        storage, path = cls.split_url(source)
-        client = cls(storage, **kwargs)
-        return client, path
-
     async def fetch(self, listing, start_prefix=""):
         data_storage = listing.data_storage.clone()
         if start_prefix:
@@ -134,13 +122,12 @@ class FSSpecClient(Client):
 
     async def _fetch_dir(self, dir_id, prefix, delimiter, pbar, listing, data_storage):
         path = f"{self.name}/{prefix}"
-        # pylint:disable-next=protected-access
-        infos = await self.fs._ls(path, detail=True, versions=True)
+        infos = await self.ls_dir(path)
         files = []
         subdirs = set()
         for info in infos:
             full_path = info["name"]
-            _, subprefix, _ = self.fs.split_path(info["name"])
+            subprefix = self.rel_path(full_path)
             if info["type"] == "directory":
                 name = full_path.split(delimiter)[-1]
                 new_dir_id = await listing.insert_dir(
@@ -155,12 +142,22 @@ class FSSpecClient(Client):
                 files.append(self._dict_from_info(info, dir_id, delimiter, subprefix))
         if files:
             await data_storage.insert_entries(files)
+            await data_storage.update_last_inserted_at()
         pbar.update(len(subdirs) + len(files))
         return subdirs
 
-    @classmethod
+    async def ls_dir(self, path):
+        # pylint:disable-next=protected-access
+        return await self.fs._ls(path, detail=True, versions=True)
+
+    def rel_path(self, path):
+        return self.fs.split_path(path)[1]
+
+    def get_full_path(self, rel_path):
+        return f"{self.PREFIX}{self.name}/{rel_path}"
+
     @abstractmethod
-    def _dict_from_info(cls, v, parent_id, delimiter, path):
+    def _dict_from_info(self, v, parent_id, delimiter, path):
         ...
 
     def fetch_nodes(
