@@ -1,4 +1,7 @@
 from base64 import b64decode
+from datetime import datetime
+
+from dateutil.parser import isoparse
 
 from dql.vendored.gcsfs import GCSFileSystem
 
@@ -14,11 +17,18 @@ class GCSClient(FSSpecClient):
     PREFIX = "gcs://"
     protocol = "gcs"
 
-    def __init__(self, name, **kwargs):
-        super().__init__(name, self.create_fs(**kwargs))
+    @staticmethod
+    def parse_timestamp(timestamp: str) -> datetime:
+        """
+        Parse timestamp string returned by GCSFileSystem.
 
-    @classmethod
-    def _dict_from_info(cls, v, parent_id, delimiter, path):
+        This ensures that the passed timestamp is timezone aware.
+        """
+        dt = isoparse(timestamp)
+        assert dt.tzinfo is not None
+        return dt
+
+    def _dict_from_info(self, v, parent_id, delimiter, path):
         name = v.get("name", "").split(delimiter)[-1]
         if "generation" in v:
             gen = f"#{v['generation']}"
@@ -34,7 +44,7 @@ class GCSClient(FSSpecClient):
             "etag": v.get("etag", ""),
             "version": v.get("generation", ""),
             "is_latest": not v.get("timeDeleted"),
-            "last_modified": v["updated"],
+            "last_modified": self.parse_timestamp(v["updated"]),
             "size": v.get("size", ""),
             # 'storage_class': v.get('StorageClass'),
             "owner_name": "",
