@@ -1,2 +1,2 @@
 from dql.data_storage.abstract import AbstractDataStorage  # noqa: F401
-from dql.data_storage.default import DefaultDataStorage  # noqa: F401
+from dql.data_storage.sqlite import SQLiteDataStorage  # noqa: F401
