@@ -44,6 +44,35 @@ class AbstractDataStorage(ABC):
     ) -> None:
         """Inserts data to shadow dataset based on bucket uri and glob path"""
 
+    @abstractmethod
+    def rename_dataset_rows(self, name: str, version: int) -> None:
+        """
+        When registering dataset, we need to rename rows table from
+        ds_<id>_shadow to ds_<id>_<version>.
+        Example: from ds_24_shadow to ds_24_1
+        """
+
+    @abstractmethod
+    def create_dataset_version(
+        self, name: str, version: int, create_rows_table=True
+    ) -> "DatasetRecord":
+        """Creates new dataset version, optionally creating new rows table"""
+
+    @abstractmethod
+    def merge_dataset_rows(
+        self,
+        src: "DatasetRecord",
+        dst: "DatasetRecord",
+        src_version: Optional[int] = None,
+        dst_version: Optional[int] = None,
+    ) -> None:
+        """
+        Merges source dataset rows and current latest destination dataset rows
+        into a new rows table created for new destination dataset version.
+        Note that table for new destination version must be created upfront.
+        Merge results should not contain duplicates.
+        """
+
     @abstractmethod
     def update_dataset(self, dataset_name: str, **kwargs) -> None:
         """Updates dataset fields"""
@@ -54,13 +83,22 @@ class AbstractDataStorage(ABC):
 
     @abstractmethod
     def get_dataset_rows(
-        self, name: str, limit: Optional[int] = None
+        self, name: str, limit: Optional[int] = None, version: Optional[int] = None
     ) -> Iterable["DatasetRow"]:
         """Gets dataset rows"""
 
     @abstractmethod
-    def remove_dataset(self, name: str) -> None:
-        """Removes dataset by name and it's corresponding entries"""
+    def remove_dataset_version(self, dataset: "DatasetRecord", version: int) -> None:
+        """
+        Deletes one single dataset version. If it was last version,
+        it removes dataset completely
+        """
+
+    @abstractmethod
+    def remove_shadow_dataset(self, dataset: "DatasetRecord") -> None:
+        """
+        Removes shadow dataset and it's corresponding rows
+        """
 
     @abstractmethod
     async def insert_entry(self, entry: Dict[str, Any]) -> int:
@@ -110,6 +148,17 @@ class AbstractDataStorage(ABC):
             - boolean saying if this storage is newly created
         """
 
+    @abstractmethod
+    async def update_last_inserted_at(self, uri: Optional[str] = None) -> None:
+        """Updates last inserted datetime in bucket with current time"""
+
+    @abstractmethod
+    def find_stale_storages(self):
+        """
+        Finds all pending storages for which the last inserted node has happened
+        before STALE_HOURS_LIMIT hours, and marks it as STALE
+        """
+
     @abstractmethod
     def mark_storage_indexed(
         self,
