@@ -1,10 +1,12 @@
 import hashlib
 import json
 import logging
+import operator
 import posixpath
 import sqlite3
-from datetime import datetime, timezone
+from datetime import MAXYEAR, MINYEAR, datetime, timezone
 from functools import partial, wraps
+from itertools import groupby
 from time import sleep
 from typing import Any, Dict, Iterable, Iterator, List, Optional, Tuple, Union
 
@@ -25,6 +27,28 @@ RETRY_MAX_TIMES = 10
 RETRY_FACTOR = 2
 
 
+def adapt_datetime(val: datetime) -> str:
+    if not (val.tzinfo is timezone.utc or val.tzname() == "UTC"):
+        try:
+            val = val.astimezone(timezone.utc)
+        except (OverflowError, ValueError, OSError):
+            if val.year == MAXYEAR:
+                val = datetime.max
+            elif val.year == MINYEAR:
+                val = datetime.min
+            else:
+                raise
+    return val.replace(tzinfo=None).isoformat(" ")
+
+
+def convert_datetime(val: bytes) -> datetime:
+    return datetime.fromisoformat(val.decode()).replace(tzinfo=timezone.utc)
+
+
+sqlite3.register_adapter(datetime, adapt_datetime)
+sqlite3.register_converter("datetime", convert_datetime)
+
+
 class NodeSchema(Schema):
     fields = [
         "id",
@@ -59,12 +83,38 @@ DATASET_FIELDS = [
     "version",
     "labels",
 ]
+
+DATASET_ROW_FIELDS = [
+    "id",
+    "dir_type",
+    "parent_id",
+    "name",
+    "checksum",
+    "etag",
+    "version",
+    "is_latest",
+    "last_modified",
+    "size",
+    "owner_name",
+    "owner_id",
+    "path_str",
+    "anno",
+    "source",
+]
+
 PATH_STR_INDEX = NodeSchema.fields.index("path_str")
 
 
 class StorageSchema(Schema):
     table = "buckets"
-    fields = ["uri", "timestamp", "expires", "status"]
+    fields = [
+        "uri",
+        "timestamp",
+        "expires",
+        "started_inserting_at",
+        "last_inserted_at",
+        "status",
+    ]
 
 
 class PartialSchema(Schema):
@@ -119,16 +169,18 @@ class SQLiteDataStorage(AbstractDataStorage):
     DATASET_TABLE_PREFIX = "ds_"
     TABLE_NAME_SHA_LIMIT = 12
 
-    def __init__(self, db_file: Optional[str] = None, table_name: str = ""):
-        self.table_name = table_name
+    def __init__(self, db_file: Optional[str] = None, uri: str = ""):
+        self.uri = uri
         self.db_file = db_file if db_file else DQLDir.find().db
-
+        detect_types = sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES
         try:
             if self.db_file == ":memory:":
                 # Enable multithreaded usage of the same in-memory db
-                self.db = sqlite3.connect("file::memory:?cache=shared", uri=True)
+                self.db = sqlite3.connect(
+                    "file::memory:?cache=shared", uri=True, detect_types=detect_types
+                )
             else:
-                self.db = sqlite3.connect(self.db_file)
+                self.db = sqlite3.connect(self.db_file, detect_types=detect_types)
             self.db.isolation_level = None  # Use autocommit mode
             self.db.execute("PRAGMA foreign_keys = ON")
             self.db.execute("PRAGMA cache_size = -102400")  # 100 MiB
@@ -138,7 +190,7 @@ class SQLiteDataStorage(AbstractDataStorage):
             self.db.execute("PRAGMA case_sensitive_like = ON")
 
             self._init_storage_table()
-            self._init_datasets_table()
+            self._init_datasets_tables()
         except RuntimeError:
             raise DQLError("Can't connect to SQLite DB")
 
@@ -151,23 +203,33 @@ class SQLiteDataStorage(AbstractDataStorage):
                 uri         TEXT PRIMARY KEY NOT NULL,
                 timestamp   DATETIME,
                 expires     DATETIME,
+                started_inserting_at  DATETIME,
+                last_inserted_at  DATETIME,
                 status      INTEGER NOT NULL
             )
         """
         )
 
-    def _init_datasets_table(self) -> None:
-        self.db.execute(
+    def _init_datasets_tables(self) -> None:
+        self.db.executescript(
             """
             CREATE TABLE IF NOT EXISTS datasets
             (
                 id              INTEGER PRIMARY KEY,
                 name            TEXT NOT NULL UNIQUE,
                 description     TEXT,
-                version         INTEGER,
                 labels          JSON DEFAULT('[]'),
                 shadow          BOOL NOT NULL
             );
+            CREATE TABLE IF NOT EXISTS datasets_versions
+            (
+                id              INTEGER PRIMARY KEY,
+                dataset_id      INTEGER NOT NULL,
+                version         INTEGER NOT NULL,
+                created_at      DATETIME,
+                FOREIGN KEY(dataset_id) REFERENCES datasets(id) ON DELETE CASCADE,
+                UNIQUE(dataset_id, version)
+            );
             """
         )
 
@@ -231,12 +293,25 @@ class SQLiteDataStorage(AbstractDataStorage):
         return cls._table_name(uri, cls.LISTING_TABLE_NAME_PREFIX)
 
     @classmethod
-    def _dataset_table_name(cls, dataset_id: int) -> str:
-        return cls.DATASET_TABLE_PREFIX + str(dataset_id)
+    def _dataset_table_name(cls, dataset_id: int, version=None) -> str:
+        name = cls.DATASET_TABLE_PREFIX + str(dataset_id)
+        if version:
+            name += f"_{version}"
+        else:
+            name += "_shadow"
+
+        return name
+
+    @property
+    def table_name(self) -> Optional[str]:
+        if not self.uri:
+            return None
+
+        return self._listing_table_name(self.uri)
 
     def clone(self, uri: Optional[str] = None) -> "SQLiteDataStorage":
-        table_name = self._listing_table_name(uri) if uri else self.table_name
-        return SQLiteDataStorage(db_file=self.db_file, table_name=table_name)
+        uri = uri or self.uri
+        return SQLiteDataStorage(db_file=self.db_file, uri=uri)
 
     # Query starters
     @property
@@ -279,10 +354,15 @@ class SQLiteDataStorage(AbstractDataStorage):
         Gets node that correspond some path list, e.g ["data-lakes", "dogs-and-cats"]
         """
         path_str = "/".join(path_list)
-        node = self.nodes.with_path().where(path_str=path_str, is_latest=True).get()
-        if not node:
+        nodes = (
+            self.nodes.with_path()
+            .where(path_str=path_str, is_latest=True)
+            .order_by("dir_type")
+            .all()
+        )
+        if not nodes:
             raise FileNotFoundError(f"Unable to resolve path {path_str}")
-        return node
+        return nodes[-1]
 
     def _populate_nodes_by_path(
         self, path_list: List[str], num: int, res: List[NodeWithPath]
@@ -339,7 +419,12 @@ class SQLiteDataStorage(AbstractDataStorage):
             else:
                 raise RuntimeError(f"No Path for node data: {d}")
 
-        d = {"name": "", "is_latest": True, "size": 0, **d}
+        d = {
+            "name": "",
+            "is_latest": True,
+            "size": 0,
+            **d,
+        }
         return {f: d.get(f) for f in NodeSchema.fields[1:]}
 
     def get_datasets(
@@ -348,29 +433,62 @@ class SQLiteDataStorage(AbstractDataStorage):
         if shadow_only is None:
             cond = ""
         elif shadow_only:
-            cond = "WHERE shadow"
+            cond = "WHERE datasets.shadow"
         else:
-            cond = "WHERE NOT shadow"
-        for row in self.db.execute(
+            cond = "WHERE NOT datasets.shadow"
+
+        rows = self.db.execute(
             f"""
             SELECT
-                id,
-                name,
-                description,
-                version,
-                labels,
-                shadow
+                datasets.id,
+                datasets.name,
+                datasets.description,
+                datasets.labels,
+                datasets.shadow,
+                datasets_versions.version
             FROM datasets
+            LEFT JOIN datasets_versions ON datasets.id = datasets_versions.dataset_id
             {cond}
             """,
-        ).fetchall():
-            yield DatasetRecord.parse(*row)
+        ).fetchall()
 
-    def get_dataset_rows(self, name: str, limit=20) -> Iterable[DatasetRow]:
-        # TODO implement listing specific versions
+        for _, g in groupby(rows, operator.itemgetter(0)):
+            dataset = self._parse_dataset(list(g))
+            if dataset:
+                yield dataset
+
+    def create_dataset_rows_table(self, name):
+        # creates special shadow dataset table to store rows
+        self.db.execute(
+            f"""
+            CREATE TABLE IF NOT EXISTS {name}
+            (
+                id              INTEGER PRIMARY KEY,
+                dir_type        INTEGER,
+                parent_id       INTEGER,
+                name            TEXT NOT NULL,
+                checksum        TEXT,
+                etag            TEXT,
+                version         TEXT,
+                is_latest       BOOL,
+                last_modified   DATETIME,
+                size            BIGINT NOT NULL,
+                owner_name      TEXT,
+                owner_id        TEXT,
+                path_str        TEXT,
+                anno            JSON,
+                source          TEXT NOT NULL,
+                UNIQUE(source, path_str)
+            );
+            """
+        )
+
+    def get_dataset_rows(
+        self, name: str, limit=20, version=None
+    ) -> Iterable[DatasetRow]:
         dataset = self.get_dataset(name)
         assert dataset
-        dataset_table_name = self._dataset_table_name(dataset.id)
+        dataset_table_name = self._dataset_table_name(dataset.id, version)
 
         limit_q = ""
         args = []
@@ -398,42 +516,21 @@ class SQLiteDataStorage(AbstractDataStorage):
             )
             dataset = self.get_dataset(name)
             assert dataset
+            assert dataset.shadow
             table_name = self._dataset_table_name(dataset.id)
-
-            # creates special shadow dataset table to store entries
-            self.db.execute(
-                f"""
-                CREATE TABLE IF NOT EXISTS {table_name}
-                (
-                    id              INTEGER PRIMARY KEY,
-                    dir_type        INTEGER,
-                    parent_id       INTEGER,
-                    name            TEXT NOT NULL,
-                    checksum        TEXT,
-                    etag            TEXT,
-                    version         TEXT,
-                    is_latest       BOOL,
-                    last_modified   DATETIME,
-                    size            BIGINT NOT NULL,
-                    owner_name      TEXT,
-                    owner_id        TEXT,
-                    path_str        TEXT,
-                    anno            JSON,
-                    source          TEXT NOT NULL
-                );
-                """
-            )
+            self.create_dataset_rows_table(table_name)
 
     def insert_into_shadow_dataset(
         self, name: str, uri: str, path: str, recursive=False
     ) -> None:
         dataset = self.get_dataset(name)
         assert dataset
+        assert dataset.shadow
 
         source_table_name = self._listing_table_name(uri)
         dataset_table_name = self._dataset_table_name(dataset.id)
 
-        self.table_name = source_table_name  # needed for parent node and .nodes
+        self.uri = uri  # needed for parent node and .nodes
         if recursive:
             if not path.endswith("*"):
                 path = path.rstrip("/") + "/*"  # glob filter must end with /*
@@ -488,6 +585,15 @@ class SQLiteDataStorage(AbstractDataStorage):
         )
         self.db.commit()
 
+    def rename_dataset_rows(self, name: str, version: int) -> None:
+        dataset = self.get_dataset(name)
+        assert dataset
+
+        old_name = self._dataset_table_name(dataset.id, None)
+        new_name = self._dataset_table_name(dataset.id, version)
+
+        self.db.execute(f"ALTER TABLE {old_name} RENAME TO {new_name}")
+
     def update_dataset(self, dataset_name: str, **kwargs) -> None:
         args = []
         fields = []
@@ -508,25 +614,110 @@ class SQLiteDataStorage(AbstractDataStorage):
         set_query = "SET " + ", ".join(f"{field} = ?" for field in fields)
         self.db.execute(f"UPDATE datasets {set_query} WHERE name = ?", args)
 
-    def get_dataset(self, name: str) -> Optional[DatasetRecord]:
-        res = self.db.execute(
-            "SELECT * from datasets WHERE name = ?", [name]
-        ).fetchall()
-        if not res:
-            return None
-
-        assert len(res) == 1, f"Dataset duplication: {name}"
-        return DatasetRecord.parse(*res[0])
-
-    def remove_dataset(self, name: str) -> None:
+    def create_dataset_version(
+        self, name: str, version: int, create_rows_table=True
+    ) -> DatasetRecord:
         with self.db:
             self.db.execute("begin")
+
             dataset = self.get_dataset(name)
             assert dataset
-            dataset_table_name = self._dataset_table_name(dataset.id)
 
+            self.db.execute(
+                """INSERT INTO datasets_versions(
+                    dataset_id, version, created_at
+                ) VALUES (?, ?, ?)
+                ON CONFLICT(dataset_id, version) DO NOTHING
+                """,
+                [dataset.id, version, datetime.now(timezone.utc)],
+            )
+
+            if create_rows_table:
+                table_name = self._dataset_table_name(dataset.id, version)
+                self.create_dataset_rows_table(table_name)
+
+            return dataset
+
+    def merge_dataset_rows(
+        self,
+        src: DatasetRecord,
+        dst: DatasetRecord,
+        src_version: Optional[int] = None,
+        dst_version: Optional[int] = None,
+    ) -> None:
+        src_table_name = self._dataset_table_name(src.id, src_version)
+        dst_table_name = self._dataset_table_name(dst.id, dst_version)
+        dst_table_name_latest = self._dataset_table_name(dst.id, dst.latest_version)
+
+        select_fields = ",".join([f for f in DATASET_ROW_FIELDS if f != "id"])
+        self.db.execute(
+            f"""
+            INSERT OR IGNORE INTO {dst_table_name} ({select_fields})
+            SELECT {select_fields} FROM {src_table_name}
+            UNION SELECT {select_fields} from {dst_table_name_latest}
+            """,
+        )
+
+    @staticmethod
+    def _parse_dataset(rows) -> Optional[DatasetRecord]:
+        dataset = None
+        for r in rows:
+            if not dataset:
+                dataset = DatasetRecord.parse(*r)
+                continue
+            dataset.merge_versions(DatasetRecord.parse(*r))  # type: ignore[unreachable]
+
+        return dataset
+
+    def get_dataset(self, name: str) -> Optional[DatasetRecord]:
+        rows = self.db.execute(
+            """
+            SELECT
+                datasets.id,
+                datasets.name,
+                datasets.description,
+                datasets.labels,
+                datasets.shadow,
+                datasets_versions.version
+            FROM datasets
+            LEFT JOIN datasets_versions ON datasets.id = datasets_versions.dataset_id
+            WHERE datasets.name = ?
+            """,
+            [name],
+        ).fetchall()
+        if not rows:
+            return None
+
+        return self._parse_dataset(rows)
+
+    def remove_dataset_version(self, dataset: DatasetRecord, version: int) -> None:
+        if not dataset.has_version(version):
+            return
+
+        self.db.execute(
+            """
+            DELETE FROM datasets_versions WHERE dataset_id = ? and version = ?
+            """,
+            [dataset.id, version],
+        )
+
+        if dataset.versions and len(dataset.versions) == 1:
+            # had only one version, fully deleting dataset
             self.db.execute("DELETE FROM datasets WHERE id = ?", [dataset.id])
-            self.db.execute(f"DROP TABLE {dataset_table_name}")
+
+        dataset.remove_version(version)
+
+        self.db.execute(f"DROP TABLE {self._dataset_table_name(dataset.id, version)}")
+
+    def remove_shadow_dataset(self, dataset: DatasetRecord) -> None:
+        self.db.execute("DELETE FROM datasets WHERE id = ?", [dataset.id])
+        self.db.execute(f"DROP TABLE {self._dataset_table_name(dataset.id)}")
+
+    @retry_sqlite_locks_async
+    async def update_last_inserted_at(self, uri: Optional[str] = None) -> None:
+        uri = uri or self.uri
+        updates = {"last_inserted_at": datetime.now(timezone.utc)}
+        self.storages.where(uri=uri).update(**updates)
 
     @retry_sqlite_locks_async
     async def insert_entry(self, entry: Dict[str, Any]) -> int:
@@ -557,6 +748,19 @@ class SQLiteDataStorage(AbstractDataStorage):
     def create_storage_if_not_registered(self, uri: str) -> None:
         self.storages.insert({"uri": uri, "status": StorageStatus.CREATED}, silent=True)
 
+    def find_stale_storages(self):
+        """
+        Finds all pending storages for which the last inserted node has happened
+        before STALE_HOURS_LIMIT hours, and marks it as STALE
+        """
+        with self.db:
+            self.db.execute("begin")
+            pending_storages = self.storages.where(status=StorageStatus.PENDING).all()
+            for storage in pending_storages:
+                if storage.is_stale:
+                    print(f"Marking storage {storage.uri} as stale")
+                    self.mark_storage_stale(storage)
+
     def register_storage_for_indexing(
         self,
         uri: str,
@@ -588,7 +792,7 @@ class SQLiteDataStorage(AbstractDataStorage):
             if storage.status == StorageStatus.PENDING:
                 return storage, False, True, False
 
-            elif storage.is_expired:
+            elif storage.is_expired or storage.status == StorageStatus.STALE:
                 storage = self.mark_storage_pending(storage)
                 return storage, True, False, False
 
@@ -629,7 +833,21 @@ class SQLiteDataStorage(AbstractDataStorage):
     @retry_sqlite_locks
     def mark_storage_pending(self, storage: Storage) -> Storage:
         # Update status to pending and dates
-        updates = {"status": StorageStatus.PENDING, "timestamp": None, "expires": None}
+        updates = {
+            "status": StorageStatus.PENDING,
+            "timestamp": None,
+            "expires": None,
+            "last_inserted_at": None,
+            "started_inserting_at": datetime.now(timezone.utc),
+        }
+        storage = storage._replace(**updates)  # type: ignore [arg-type]
+        self.storages.where(uri=storage.uri).update(**updates)
+        return storage
+
+    @retry_sqlite_locks
+    def mark_storage_stale(self, storage: Storage) -> Storage:
+        # Update status to pending and dates
+        updates = {"status": StorageStatus.STALE, "timestamp": None, "expires": None}
         storage = storage._replace(**updates)  # type: ignore [arg-type]
         self.storages.where(uri=storage.uri).update(**updates)
         return storage
@@ -657,6 +875,7 @@ class SQLiteDataStorage(AbstractDataStorage):
                 timestamp=end_time,
                 expires=expires,
                 status=status,
+                last_inserted_at=end_time,
             )
 
             if not self.table_name:
