@@ -2,10 +2,12 @@ import os
 from collections import defaultdict
 from itertools import zip_longest
 from shutil import copy2
-from typing import DefaultDict, Iterable, List
+from typing import DefaultDict, List
 
 import yaml
+from dvc_data.hashfile.db.local import LocalHashFileDB
 from fsspec.asyn import get_loop, sync
+from funcy import notnone, select_values
 from reflink import reflink
 from reflink.error import ReflinkImpossibleError
 from tqdm import tqdm
@@ -15,14 +17,37 @@ from dql.data_storage import AbstractDataStorage
 from dql.node import DirType, Node, NodeWithPath
 from dql.nodes_fetcher import JsonNodesFetcher
 from dql.storage import Storage
+from dql.utils import suffix_to_number
 
 
 def check_checksums(nodes):
     for node in nodes:
         if node.name and not node.checksum:
-            raise ValueError(
-                f"Instantiation Error: Missing checksum for node: {node}"
-            )
+            raise ValueError(f"Instantiation Error: Missing checksum for node: {node}")
+
+
+def instantiate_node(
+    node: NodeWithPath,
+    cache: LocalHashFileDB,
+    output: str,
+    progress_bar: tqdm,
+    force: bool = False,
+):
+    if not node.name:
+        return
+    src = cache.oid_to_path(node.checksum)  # type: ignore[attr-defined]
+    dst = os.path.join(output, *node.path)  # type: ignore[attr-defined]
+    if os.path.exists(dst):
+        if force:
+            os.remove(dst)
+        else:
+            progress_bar.close()
+            raise FileExistsError(f"Path {dst} already exists")
+    try:
+        reflink(src, dst)
+    except (NotImplementedError, ReflinkImpossibleError):
+        # Default to copy if reflinks are not supported
+        copy2(src, dst)
 
 
 class Listing:
@@ -40,8 +65,8 @@ class Listing:
     def id(self):
         return self.storage.id
 
-    def fetch(self):
-        sync(get_loop(), self.client.fetch, self)
+    def fetch(self, start_prefix=""):
+        sync(get_loop(), self.client.fetch, self, start_prefix)
 
     @staticmethod
     async def _insert_dir(data_storage, parent_id, name, time, path):
@@ -83,15 +108,13 @@ class Listing:
         await self.insert_file_from_node(node)
 
     async def insert_file_from_node(self, node):
-        await self.data_storage.insert_entry(
-            {**node._asdict(), "is_dir": False}
-        )
+        await self.data_storage.insert_entry({**node._asdict(), "is_dir": False})
 
     async def insert_root(self, data_storage=None) -> int:
         return await (data_storage or self.data_storage).insert_root()
 
-    def expand_path(self, path) -> Iterable[NodeWithPath]:
-        return self.data_storage.get_nodes_by_path(path)
+    def expand_path(self, path) -> List[NodeWithPath]:
+        return self.data_storage.expand_path(path)
 
     def resolve_path(self, path) -> NodeWithPath:
         return self.data_storage.get_node_by_path(path)
@@ -105,7 +128,7 @@ class Listing:
         node,
         dataset_file,
     ):
-        nodes = self.data_storage.walk_subtree(node, sort_by_size=True)
+        nodes = self.data_storage.walk_subtree(node, sort="-size")
 
         print(f"Creating '{dataset_file}'")
 
@@ -132,8 +155,10 @@ class Listing:
         total_files=None,
         recursive=False,
         copy_dir_contents=False,
+        force=False,
         virtual_only=False,
         relative_path=None,
+        shared_progress_bar=None,
     ):
         rel_path_elements = relative_path.split("/") if relative_path else []
         all_nodes = []
@@ -150,9 +175,8 @@ class Listing:
                     dir_path.append(node.name)
                 subtree_nodes = self.data_storage.walk_subtree(
                     node,
-                    sort_by_parent_id=True,
-                    sort_by_path=True,
-                    type_filter="files",
+                    sort=["parent_id", "path_str"],
+                    type="files",
                 )
                 all_nodes.extend(
                     n._replace(path=dir_path + n.path) for n in subtree_nodes
@@ -167,7 +191,7 @@ class Listing:
 
         nodes = iter(all_nodes)
 
-        progress_bar = tqdm(
+        progress_bar = shared_progress_bar or tqdm(
             desc=f"Instantiating '{output}'",
             unit=" files",
             unit_scale=True,
@@ -187,14 +211,7 @@ class Listing:
             os.makedirs(curr_dir_path_str, exist_ok=True)
 
             while node and node.parent_id == curr_parent_id:
-                if node.name:
-                    src = cache.oid_to_path(node.checksum)
-                    dst = os.path.join(output, *node.path)
-                    try:
-                        reflink(src, dst)
-                    except (NotImplementedError, ReflinkImpossibleError):
-                        # Default to copy if reflinks are not supported
-                        copy2(src, dst)
+                instantiate_node(node, cache, output, progress_bar, force)
 
                 counter += 1
                 if counter > 1000:
@@ -242,18 +259,17 @@ class Listing:
         cache,
         total_size=None,
         recursive=False,
+        shared_progress_bar=None,
     ):
         nodes_by_path: DefaultDict[str, List] = defaultdict(list)
         for node in nodes:
             node_path = "/".join(node.path)
             if recursive and node.is_dir:
                 nodes_by_path[node_path].extend(
-                    self.data_storage.walk_subtree(node, sort_by_size=True)
+                    self.data_storage.walk_subtree(node, sort="-size")
                 )
             else:
-                nodes_by_path[node_path].append(
-                    node._replace(path=[node.name])
-                )
+                nodes_by_path[node_path].append(node._replace(path=[node.name]))
 
         updated_node_lookup = {}
         for group_path, all_nodes in nodes_by_path.items():
@@ -263,6 +279,7 @@ class Listing:
                 cache,
                 self.data_storage,
                 total_size,
+                shared_progress_bar=shared_progress_bar,
             )
             # This assert is hopefully not necessary, but is here mostly to
             # ensure this functionality stays consistent (and catch any bugs)
@@ -273,9 +290,7 @@ class Listing:
             else:
                 group_path_elements = []
             for node in updated_nodes:
-                updated_node_lookup[node.id] = node._replace(
-                    path=group_path_elements
-                )
+                updated_node_lookup[node.id] = node._replace(path=group_path_elements)
 
         # Since nodes can have updated information (such as checksums)
         # after they are downloaded
@@ -309,16 +324,25 @@ class Listing:
         type=None,  # pylint: disable=redefined-builtin
         jmespath=None,
     ):
-        return self.data_storage.find(
-            node,
-            names=names,
-            inames=inames,
-            paths=paths,
-            ipaths=ipaths,
-            size=size,
-            type=type,
-            jmespath=jmespath,
+        conds = select_values(
+            notnone,
+            {
+                "name__glob": names,
+                "name__iglob": inames,
+                "path_str__glob": paths,
+                "path_str__iglob": ipaths,
+                "type": type,
+            },
         )
 
+        if size is not None:
+            size_limit = suffix_to_number(size)
+            if size_limit >= 0:
+                conds["size__gte"] = size_limit
+            else:
+                conds["size__lte"] = -size_limit
+
+        return self.data_storage.find(node, jmespath=jmespath, **conds)
+
     def du(self, node):
         return self.data_storage.size(node)
