@@ -1,12 +1,14 @@
 import posixpath
 from abc import ABC, abstractmethod
-from datetime import datetime, timedelta
+from datetime import datetime, timedelta, timezone
 from functools import cached_property
 from typing import NamedTuple, Optional, Union
 from urllib.parse import urlparse
 
 from dql.utils import is_expired, time_to_local_str, time_to_str
 
+STALE_HOURS_LIMIT = 4
+
 
 class Status:
     CREATED = 1
@@ -14,6 +16,7 @@ class Status:
     FAILED = 3
     COMPLETE = 4
     PARTIAL = 5
+    STALE = 6
 
 
 class AbstractStorage(ABC):
@@ -54,6 +57,8 @@ class StorageRecord(NamedTuple):
     uri: str
     timestamp: Optional[Union[datetime, str]] = None
     expires: Optional[Union[datetime, str]] = None
+    started_inserting_at: Optional[Union[datetime, str]] = None
+    last_inserted_at: Optional[Union[datetime, str]] = None
     status: int = Status.CREATED
 
 
@@ -62,6 +67,17 @@ class Storage(StorageRecord, AbstractStorage):
     def is_expired(self) -> bool:
         return is_expired(self.expires)
 
+    @property
+    def is_pending(self) -> bool:
+        return self.status == Status.PENDING
+
+    @property
+    def is_stale(self) -> bool:
+        limit = datetime.now(timezone.utc) - timedelta(hours=STALE_HOURS_LIMIT)
+        date_to_check = self.last_inserted_at or self.started_inserting_at
+
+        return self.is_pending and date_to_check < limit  # type: ignore [operator]
+
     @property
     def timestamp_str(self) -> Optional[str]:
         if not self.timestamp:
