@@ -1,7 +1,8 @@
 import os
 from pathlib import PosixPath
-from typing import Dict, Generator, NamedTuple, Type
+from typing import Dict, Generator, Type
 
+import attrs
 import pytest
 from pytest import MonkeyPatch, TempPathFactory
 from upath.implementations.cloud import CloudPath
@@ -55,9 +56,14 @@ def data_storage():
     _data_storage = SQLiteDataStorage(db_file=":memory:")
     yield _data_storage
 
-    # Check that the DB isn't locked after the test
+    # Wipe the DB after the test
+    # Using new connection to check that the DB isn't locked
     conn = _data_storage.clone().db
-    conn.execute("BEGIN IMMEDIATE")
+    tables = conn.execute(
+        "SELECT name FROM sqlite_master WHERE type='table'"
+    ).fetchall()
+    for (table,) in tables:
+        conn.execute(f"DROP TABLE '{table}'")
     conn.commit()
 
 
@@ -118,7 +124,9 @@ def pytest_configure(config):
     config.dql_config.dql_git_repo = config.getoption("--dql-git-repo")
 
 
-class CloudTestServer(NamedTuple):
+@attrs.define
+class CloudServer:
+    kind: str
     src: CloudPath
     client_config: Dict[str, str]
 
@@ -133,27 +141,46 @@ def make_cloud_server(src_path, cloud_type):
         client_config = {"endpoint_url": endpoint_url}
     elif cloud_type == "azure":
         client_config = fs.storage_options.copy()
+    elif cloud_type == "file":
+        client_config = {}
     else:
         raise ValueError(f"invalid cloud_type: {cloud_type}")
     (src_path / "description").write_text("Cats and Dogs")
+    (src_path / "cats").mkdir()
     (src_path / "cats" / "cat1").write_text("meow")
     (src_path / "cats" / "cat2").write_text("mrow")
+    (src_path / "dogs").mkdir()
     (src_path / "dogs" / "dog1").write_text("woof")
     (src_path / "dogs" / "dog2").write_text("arf")
     (src_path / "dogs" / "dog3").write_text("bark")
+    (src_path / "dogs" / "others").mkdir()
     (src_path / "dogs" / "others" / "dog4").write_text("ruff")
 
-    return CloudTestServer(src=src_path, client_config=client_config)
+    return CloudServer(kind=cloud_type, src=src_path, client_config=client_config)
 
 
-class CloudTestCatalog(NamedTuple):
-    src: CloudPath
+@attrs.define
+class CloudTestCatalog:
+    server: CloudServer
     working_dir: PosixPath
     catalog: Catalog
-    client_config: Dict[str, str]
+
+    @property
+    def src(self):
+        return self.server.src
+
+    @property
+    def src_uri(self):
+        if self.server.kind == "file":
+            return self.src.as_uri()
+        return str(self.src).rstrip("/")
+
+    @property
+    def client_config(self):
+        return self.server.client_config
 
 
-@pytest.fixture(scope="session", params=["s3", "gcs", "azure"])
+@pytest.fixture(scope="session", params=["file", "s3", "gcs", "azure"])
 def cloud_type(request):
     return request.param
 
@@ -170,6 +197,11 @@ def cloud_server(request, tmp_upath_factory, cloud_type, version_aware):
             src_path = tmp_upath_factory.azure(conn_str)
         else:
             pytest.skip("Can't test versioning with Azure")
+    elif cloud_type == "file":
+        if version_aware:
+            pytest.skip("Local storage can't be versioned")
+        else:
+            src_path = tmp_upath_factory.mktemp("local")
     else:
         src_path = tmp_upath_factory.mktemp(cloud_type, version_aware=version_aware)
     return make_cloud_server(src_path, cloud_type)
@@ -187,10 +219,9 @@ def cloud_test_catalog(
     tmpfile_dir.mkdir()
 
     catalog = Catalog(data_storage, cache_dir=str(cache_dir), tmp_dir=str(tmpfile_dir))
-
-    return CloudTestCatalog(
-        src=cloud_server.src,
-        working_dir=tmp_path,
-        catalog=catalog,
-        client_config=cloud_server.client_config,
+    result = CloudTestCatalog(
+        server=cloud_server, working_dir=tmp_path, catalog=catalog
     )
+    if cloud_server.kind == "file":
+        catalog.add_storage(f"{result.src_uri}")
+    return result
