@@ -1,9 +1,11 @@
+import os
 import shutil
 import sqlite3
 import uuid
 
 import pytest
 import yaml
+from fsspec.implementations.local import LocalFileSystem
 
 from dql.catalog import parse_dql_file
 from dql.utils import remove_readonly
@@ -12,20 +14,20 @@ from dql.utils import remove_readonly
 @pytest.fixture
 def dogs_shadow_dataset(cloud_test_catalog):
     shadow_dataset_name = uuid.uuid4().hex
-    src = cloud_test_catalog.src
+    src_uri = cloud_test_catalog.src_uri
     catalog = cloud_test_catalog.catalog
 
     # list source to have it in db as source for dataset
     list(
         catalog.ls(
-            [str(src)],
+            [src_uri],
             client_config=cloud_test_catalog.client_config,
         )
     )
 
     catalog.upsert_shadow_dataset(
         shadow_dataset_name,
-        [str(src / "dogs" / "*")],
+        [f"{src_uri}/dogs/*"],
         client_config=cloud_test_catalog.client_config,
         recursive=True,
     )
@@ -46,98 +48,95 @@ def dogs_registered_dataset(cloud_test_catalog, dogs_shadow_dataset):
 
 
 def test_find(cloud_test_catalog):
-    src = cloud_test_catalog.src
+    src_uri = cloud_test_catalog.src_uri
     catalog = cloud_test_catalog.catalog
     config = cloud_test_catalog.client_config
 
-    assert set(catalog.find([str(src)], client_config=config)) == {
-        str(src / "description"),
-        str(src / "cats") + "/",
-        str(src / "cats" / "cat1"),
-        str(src / "cats" / "cat2"),
-        str(src / "dogs") + "/",
-        str(src / "dogs" / "dog1"),
-        str(src / "dogs" / "dog2"),
-        str(src / "dogs" / "dog3"),
-        str(src / "dogs" / "others") + "/",
-        str(src / "dogs" / "others" / "dog4"),
+    assert set(catalog.find([src_uri], client_config=config)) == {
+        f"{src_uri}/description",
+        f"{src_uri}/cats/",
+        f"{src_uri}/cats/cat1",
+        f"{src_uri}/cats/cat2",
+        f"{src_uri}/dogs/",
+        f"{src_uri}/dogs/dog1",
+        f"{src_uri}/dogs/dog2",
+        f"{src_uri}/dogs/dog3",
+        f"{src_uri}/dogs/others/",
+        f"{src_uri}/dogs/others/dog4",
     }
 
     with pytest.raises(FileNotFoundError):
         set(
             catalog.find(
-                [str(src / "does_not_exist")],
+                [f"{src_uri}/does_not_exist"],
                 client_config=config,
             )
         )
 
 
 def test_find_names_paths_size_type(cloud_test_catalog):
-    src = cloud_test_catalog.src
+    src_uri = cloud_test_catalog.src_uri
     catalog = cloud_test_catalog.catalog
     config = cloud_test_catalog.client_config
 
-    assert set(catalog.find([str(src)], names=["*cat*"], client_config=config)) == {
-        str(src / "cats") + "/",
-        str(src / "cats" / "cat1"),
-        str(src / "cats" / "cat2"),
+    assert set(catalog.find([src_uri], names=["*cat*"], client_config=config)) == {
+        f"{src_uri}/cats/",
+        f"{src_uri}/cats/cat1",
+        f"{src_uri}/cats/cat2",
     }
 
     assert set(
-        catalog.find([str(src)], names=["*cat*"], typ="dir", client_config=config)
+        catalog.find([src_uri], names=["*cat*"], typ="dir", client_config=config)
     ) == {
-        str(src / "cats") + "/",
+        f"{src_uri}/cats/",
     }
 
     assert (
-        len(list(catalog.find([str(src)], names=["*CAT*"], client_config=config))) == 0
+        len(list(catalog.find([src_uri], names=["*CAT*"], client_config=config))) == 0
     )
 
-    assert set(catalog.find([str(src)], inames=["*CAT*"], client_config=config)) == {
-        str(src / "cats") + "/",
-        str(src / "cats" / "cat1"),
-        str(src / "cats" / "cat2"),
+    assert set(catalog.find([src_uri], inames=["*CAT*"], client_config=config)) == {
+        f"{src_uri}/cats/",
+        f"{src_uri}/cats/cat1",
+        f"{src_uri}/cats/cat2",
     }
 
-    assert set(
-        catalog.find([str(src)], paths=["*cats/cat*"], client_config=config)
-    ) == {
-        str(src / "cats" / "cat1"),
-        str(src / "cats" / "cat2"),
+    assert set(catalog.find([src_uri], paths=["*cats/cat*"], client_config=config)) == {
+        f"{src_uri}/cats/cat1",
+        f"{src_uri}/cats/cat2",
     }
 
     assert (
-        len(list(catalog.find([str(src)], paths=["*caTS/CaT**"], client_config=config)))
+        len(list(catalog.find([src_uri], paths=["*caTS/CaT**"], client_config=config)))
         == 0
     )
 
     assert set(
-        catalog.find([str(src)], ipaths=["*caTS/CaT*"], client_config=config)
+        catalog.find([src_uri], ipaths=["*caTS/CaT*"], client_config=config)
     ) == {
-        str(src / "cats" / "cat1"),
-        str(src / "cats" / "cat2"),
+        f"{src_uri}/cats/cat1",
+        f"{src_uri}/cats/cat2",
     }
 
-    assert set(catalog.find([str(src)], size="5", typ="f", client_config=config)) == {
-        str(src / "description"),
+    assert set(catalog.find([src_uri], size="5", typ="f", client_config=config)) == {
+        f"{src_uri}/description",
     }
 
-    assert set(catalog.find([str(src)], size="-3", typ="f", client_config=config)) == {
-        str(src / "dogs" / "dog2"),
+    assert set(catalog.find([src_uri], size="-3", typ="f", client_config=config)) == {
+        f"{src_uri}/dogs/dog2",
     }
 
 
-def test_find_names_columns(cloud_test_catalog):
-    src = cloud_test_catalog.src
+def test_find_names_columns(cloud_test_catalog, cloud_type):
+    src_uri = cloud_test_catalog.src_uri
     catalog = cloud_test_catalog.catalog
     config = cloud_test_catalog.client_config
 
-    src_str = str(src)
-    owner = "webfile" if src_str.startswith("s3") else ""
+    owner = "webfile" if cloud_type == "s3" else ""
 
     assert set(
         catalog.find(
-            [src_str],
+            [src_uri],
             names=["*cat*"],
             columns=["du", "name", "owner", "path", "size", "type"],
             client_config=config,
@@ -145,9 +144,9 @@ def test_find_names_columns(cloud_test_catalog):
     ) == {
         "\t".join(columns)
         for columns in [
-            ["8", "cats", "", str(src / "cats") + "/", "0", "d"],
-            ["4", "cat1", owner, str(src / "cats" / "cat1"), "4", "f"],
-            ["4", "cat2", owner, str(src / "cats" / "cat2"), "4", "f"],
+            ["8", "cats", "", f"{src_uri}/cats/", "0", "d"],
+            ["4", "cat1", owner, f"{src_uri}/cats/cat1", "4", "f"],
+            ["4", "cat2", owner, f"{src_uri}/cats/cat2", "4", "f"],
         ]
     }
 
@@ -163,16 +162,16 @@ def test_find_names_columns(cloud_test_catalog):
     ),
 )
 def test_cp_root(cloud_test_catalog, recursive, star, dir_exists):
-    src = cloud_test_catalog.src
+    src_uri = cloud_test_catalog.src_uri
     working_dir = cloud_test_catalog.working_dir
     catalog = cloud_test_catalog.catalog
 
     dest = working_dir / "data"
 
     if star:
-        src_path = f"{str(src).rstrip('/')}/*"
+        src_path = f"{src_uri}/*"
     else:
-        src_path = str(src)
+        src_path = src_uri
 
     if star:
         with pytest.raises(FileNotFoundError):
@@ -267,18 +266,18 @@ def test_cp_root(cloud_test_catalog, recursive, star, dir_exists):
     ),
 )
 def test_cp_subdir(cloud_test_catalog, recursive, star, slash, dir_exists):
-    src = cloud_test_catalog.src / "dogs"
+    src_uri = f"{cloud_test_catalog.src_uri}/dogs"
     working_dir = cloud_test_catalog.working_dir
     catalog = cloud_test_catalog.catalog
 
     dest = working_dir / "data"
 
     if star:
-        src_path = f"{str(src)}/*"
+        src_path = f"{src_uri}/*"
+    elif slash:
+        src_path = f"{src_uri}/"
     else:
-        src_path = str(src)
-        if slash:
-            src_path = f"{src_path}/"
+        src_path = src_uri
 
     if star:
         with pytest.raises(FileNotFoundError):
@@ -370,8 +369,8 @@ def test_cp_subdir(cloud_test_catalog, recursive, star, slash, dir_exists):
 )
 def test_cp_multi_subdir(cloud_test_catalog, recursive, star, slash):
     sources = [
-        cloud_test_catalog.src / "cats",
-        cloud_test_catalog.src / "dogs",
+        f"{cloud_test_catalog.src_uri}/cats",
+        f"{cloud_test_catalog.src_uri}/dogs",
     ]
     working_dir = cloud_test_catalog.working_dir
     catalog = cloud_test_catalog.catalog
@@ -379,11 +378,11 @@ def test_cp_multi_subdir(cloud_test_catalog, recursive, star, slash):
     dest = working_dir / "data"
 
     if star:
-        src_paths = [f"{str(src)}/*" for src in sources]
+        src_paths = [f"{src}/*" for src in sources]
+    elif slash:
+        src_paths = [f"{src}/" for src in sources]
     else:
-        src_paths = [str(src) for src in sources]
-        if slash:
-            src_paths = [f"{src}/" for src in src_paths]
+        src_paths = sources
 
     with pytest.raises(FileNotFoundError):
         catalog.cp(
@@ -480,14 +479,12 @@ def test_cp_multi_subdir(cloud_test_catalog, recursive, star, slash):
 
 
 def test_cp_double_subdir(cloud_test_catalog):
-    src = cloud_test_catalog.src / "dogs" / "others"
+    src_path = f"{cloud_test_catalog.src_uri}/dogs/others"
     working_dir = cloud_test_catalog.working_dir
     catalog = cloud_test_catalog.catalog
 
     dest = working_dir / "data"
 
-    src_path = f"{str(src)}/"
-
     catalog.cp(
         [src_path],
         str(dest),
@@ -520,15 +517,12 @@ def test_cp_double_subdir(cloud_test_catalog):
 
 @pytest.mark.parametrize("no_glob", (True, False))
 def test_cp_single_file(cloud_test_catalog, no_glob):
-    src = cloud_test_catalog.src / "dogs" / "dog1"
     working_dir = cloud_test_catalog.working_dir
     catalog = cloud_test_catalog.catalog
 
     dest = working_dir / "data"
 
-    src_path = str(src)
-
-    print(src_path)
+    src_path = f"{cloud_test_catalog.src_uri}/dogs/dog1"
 
     dest.mkdir()
 
@@ -547,13 +541,12 @@ def test_cp_single_file(cloud_test_catalog, no_glob):
 
 
 def test_cp_dql_file_options(cloud_test_catalog):
-    src = cloud_test_catalog.src / "dogs"
     working_dir = cloud_test_catalog.working_dir
     catalog = cloud_test_catalog.catalog
 
     dest = working_dir / "data"
 
-    src_path = str(src / "*")
+    src_path = f"{cloud_test_catalog.src_uri}/dogs/*"
 
     dql_file = working_dir / "custom_name.dql"
 
@@ -578,7 +571,7 @@ def test_cp_dql_file_options(cloud_test_catalog):
     dql_contents = yaml.safe_load(dql_file.read_text())
     assert len(dql_contents) == 1
     data = dql_contents[0]
-    assert data["data-source"]["uri"] == src_path.rstrip("/")
+    assert data["data-source"]["uri"] == src_path
     expected_file_count = 3
     assert len(data["files"]) == expected_file_count
     files_by_name = {f["name"]: f for f in data["files"]}
@@ -626,8 +619,8 @@ def test_cp_dql_file_options(cloud_test_catalog):
 
 def test_cp_dql_file_sources(cloud_test_catalog):
     sources = [
-        f"{str(cloud_test_catalog.src / 'cats')}/",
-        str(cloud_test_catalog.src / "dogs" / "*"),
+        f"{cloud_test_catalog.src_uri}/cats/",
+        f"{cloud_test_catalog.src_uri}/dogs/*",
     ]
     working_dir = cloud_test_catalog.working_dir
     catalog = cloud_test_catalog.catalog
@@ -756,12 +749,12 @@ def test_cp_dql_file_sources(cloud_test_catalog):
 
 
 def test_get(cloud_test_catalog):
-    src = cloud_test_catalog.src
+    src_uri = cloud_test_catalog.src_uri
     catalog = cloud_test_catalog.catalog
     config = cloud_test_catalog.client_config
     dest = cloud_test_catalog.working_dir / "data"
 
-    catalog.get(str(src), str(dest), client_config=config)
+    catalog.get(src_uri, str(dest), client_config=config)
 
     assert (dest / "cats" / "cat1").read_text() == "meow"
     assert (dest / "cats" / "cat2").read_text() == "mrow"
@@ -773,7 +766,7 @@ def test_get(cloud_test_catalog):
 
 
 def test_get_subdir(cloud_test_catalog):
-    src = f"{str(cloud_test_catalog.src)}/dogs"
+    src = f"{cloud_test_catalog.src_uri}/dogs"
     working_dir = cloud_test_catalog.working_dir
     catalog = cloud_test_catalog.catalog
 
@@ -805,58 +798,62 @@ def test_get_subdir(cloud_test_catalog):
 
 
 def test_du(cloud_test_catalog):
-    src = str(cloud_test_catalog.src)
+    src_uri = cloud_test_catalog.src_uri
     catalog = cloud_test_catalog.catalog
 
     expected_results = [
-        (f"{src}cats/", 8),
-        (f"{src}dogs/others/", 4),
-        (f"{src}dogs/", 15),
-        (src, 36),
+        (f"{src_uri}/cats/", 8),
+        (f"{src_uri}/dogs/others/", 4),
+        (f"{src_uri}/dogs/", 15),
+        (f"{src_uri}/", 36),
     ]
 
-    results = catalog.du([src], client_config=cloud_test_catalog.client_config)
+    results = catalog.du([src_uri], client_config=cloud_test_catalog.client_config)
 
-    assert list(results) == expected_results[3:]
+    assert set(results) == set(expected_results[3:])
 
-    results = catalog.du([src], client_config=cloud_test_catalog.client_config, depth=1)
+    results = catalog.du(
+        [src_uri], client_config=cloud_test_catalog.client_config, depth=1
+    )
 
-    assert list(results) == expected_results[:1] + expected_results[2:]
+    assert set(results) == set(expected_results[:1] + expected_results[2:])
 
-    results = catalog.du([src], client_config=cloud_test_catalog.client_config, depth=5)
+    results = catalog.du(
+        [src_uri], client_config=cloud_test_catalog.client_config, depth=5
+    )
 
-    assert list(results) == expected_results
+    assert set(results) == set(expected_results)
 
 
 def test_ls_glob(cloud_test_catalog):
-    src = cloud_test_catalog.src
+    src_uri = cloud_test_catalog.src_uri
     catalog = cloud_test_catalog.catalog
 
-    assert [
+    assert sorted(
         (source.node.name, [n.name for n in nodes])
         for source, nodes in catalog.ls(
-            [str(src / "dogs" / "dog*")],
+            [f"{src_uri}/dogs/dog*"],
             client_config=cloud_test_catalog.client_config,
         )
-    ] == [("dog1", ["dog1"]), ("dog2", ["dog2"]), ("dog3", ["dog3"])]
+    ) == [("dog1", ["dog1"]), ("dog2", ["dog2"]), ("dog3", ["dog3"])]
 
 
 def test_upserting_shadow_dataset(cloud_test_catalog):
     shadow_dataset_name = uuid.uuid4().hex
-    src = cloud_test_catalog.src
+    src_uri = cloud_test_catalog.src_uri
     catalog = cloud_test_catalog.catalog
 
     # list source to have it in db as source for dataset
     list(
         catalog.ls(
-            [str(src)],
+            [src_uri],
             client_config=cloud_test_catalog.client_config,
         )
     )
 
     catalog.upsert_shadow_dataset(
         shadow_dataset_name,
-        [str(src / "dogs" / "*")],
+        [f"{src_uri}/dogs/*"],
         client_config=cloud_test_catalog.client_config,
         recursive=True,
     )
@@ -864,7 +861,7 @@ def test_upserting_shadow_dataset(cloud_test_catalog):
     dataset = catalog.data_storage.get_dataset(shadow_dataset_name)
     assert dataset.name == shadow_dataset_name
     assert dataset.description is None
-    assert dataset.version is None
+    assert dataset.versions is None
     assert dataset.labels == []
     assert dataset.shadow is True
 
@@ -885,10 +882,10 @@ def test_registering_dataset(cloud_test_catalog, dogs_shadow_dataset):
     )
 
     dataset = catalog.data_storage.get_dataset(dogs_shadow_dataset.name)
-    dataset_table_name = catalog.data_storage._dataset_table_name(dataset.id)
+    dataset_table_name = catalog.data_storage._dataset_table_name(dataset.id, 1)
     assert dataset.name == dogs_shadow_dataset.name
     assert dataset.description == "dogs dataset"
-    assert dataset.version == 1
+    assert dataset.versions == [1]
     assert dataset.labels == ["dogs", "dataset"]
     assert dataset.shadow is False
     data = catalog.data_storage.db.execute(
@@ -909,7 +906,7 @@ def test_registering_dataset_with_new_name(cloud_test_catalog, dogs_shadow_datas
     )
     dataset = catalog.data_storage.get_dataset(new_dataset_name)
     assert dataset
-    dataset_table_name = catalog.data_storage._dataset_table_name(dataset.id)
+    dataset_table_name = catalog.data_storage._dataset_table_name(dataset.id, 1)
     assert dataset.name == new_dataset_name
     data = catalog.data_storage.db.execute(
         f"select * from {dataset_table_name}"
@@ -930,7 +927,7 @@ def test_registering_dataset_with_custom_version(
     )
 
     dataset = catalog.data_storage.get_dataset(dogs_shadow_dataset.name)
-    assert dataset.version == 5
+    assert dataset.versions == [5]
 
 
 def test_removing_dataset(cloud_test_catalog, dogs_shadow_dataset):
@@ -965,7 +962,7 @@ def test_edit_dataset(cloud_test_catalog, dogs_registered_dataset):
     )
 
     dataset = catalog.data_storage.get_dataset(dataset_new_name)
-    assert dataset.version == 1
+    assert dataset.versions == [1]
     assert dataset.name == dataset_new_name
     assert dataset.description == "new description"
     assert dataset.labels == ["cats", "birds"]
@@ -974,9 +971,41 @@ def test_edit_dataset(cloud_test_catalog, dogs_registered_dataset):
 def test_ls_dataset_rows(cloud_test_catalog, dogs_registered_dataset):
     catalog = cloud_test_catalog.catalog
 
-    assert [r.name for r in catalog.ls_dataset_rows(dogs_registered_dataset.name)] == [
+    assert {
+        r.name for r in catalog.ls_dataset_rows(dogs_registered_dataset.name, version=1)
+    } == {
         "dog1",
         "dog2",
         "dog3",
         "dog4",
-    ]
+    }
+
+
+def clear_storages(catalog):
+    catalog.data_storage.storages.delete()
+
+
+@pytest.mark.parametrize("cloud_type", ["file"], indirect=True)
+@pytest.mark.parametrize("use_path", [False, True])
+def test_add_storage(cloud_test_catalog, use_path):
+    catalog = cloud_test_catalog.catalog
+    src_uri = cloud_test_catalog.src_uri
+    clear_storages(catalog)
+    if use_path:
+        src = LocalFileSystem._strip_protocol(src_uri)
+        src = os.path.normpath(src)
+    else:
+        src = src_uri
+    with pytest.raises(RuntimeError):
+        catalog.index(src)
+    catalog.add_storage(src)
+    assert catalog.index([src_uri])
+
+
+def test_add_storage_error(cloud_test_catalog, cloud_type):
+    catalog = cloud_test_catalog.catalog
+    src_uri = cloud_test_catalog.src_uri
+    if cloud_type == "file":
+        src_uri += "/invalid"
+    with pytest.raises(RuntimeError):
+        catalog.add_storage(src_uri)
