@@ -43,9 +43,7 @@ def make_size_node(node_id, dir_type, parent_id, name, size):
 
 
 def test_node_bucket_the_only_item():
-    bkt = NodeChunk(
-        iter([make_size_node(20, DirType.FILE, 2, "file.csv", 100)]), 201
-    )
+    bkt = NodeChunk(iter([make_size_node(20, DirType.FILE, 2, "file.csv", 100)]), 201)
 
     result = next(bkt)
     assert len(result) == 1
@@ -53,9 +51,7 @@ def test_node_bucket_the_only_item():
 
 
 def test_node_bucket_the_only_item_over_limit():
-    bkt = NodeChunk(
-        iter([make_size_node(20, DirType.FILE, 2, "file.csv", 100)]), 1
-    )
+    bkt = NodeChunk(iter([make_size_node(20, DirType.FILE, 2, "file.csv", 100)]), 1)
 
     result = next(bkt)
     assert len(result) == 1
@@ -63,9 +59,7 @@ def test_node_bucket_the_only_item_over_limit():
 
 
 def test_node_bucket_the_last_one():
-    bkt = NodeChunk(
-        iter([make_size_node(20, DirType.FILE, 2, "file.csv", 100)]), 1
-    )
+    bkt = NodeChunk(iter([make_size_node(20, DirType.FILE, 2, "file.csv", 100)]), 1)
 
     next(bkt)
     with pytest.raises(StopIteration):
