@@ -1,6 +1,6 @@
 import pytest
 
-from dql.utils import dql_paths_join, sizeof_fmt, sql_str, suffix_to_number
+from dql.utils import dql_paths_join, sizeof_fmt, suffix_to_number
 
 DQL_TEST_PATHS = ["/file1", "file2", "/dir/file3", "dir/file4"]
 DQL_EX_ROOT = ["/file1", "/file2", "/dir/file3", "/dir/file4"]
@@ -68,18 +68,6 @@ def test_sizeof_fmt(num, suffix, si, expected):
 @pytest.mark.parametrize(
     "text,expected",
     (
-        ("test string", "'test string'"),
-        ("Can't escape this", "'Can''t escape this'"),
-    ),
-)
-def test_sql_str(text, expected):
-    assert sql_str(text) == expected
-
-
-@pytest.mark.parametrize(
-    "text,expected",
-    (
-        ("", None),
         ("1", 1),
         ("50", 50),
         ("1K", 1024),
@@ -94,6 +82,7 @@ def test_suffix_to_number(text, expected):
 @pytest.mark.parametrize(
     "text",
     (
+        "",
         "Bogus",
         "50H",
     ),
